package examples;

/* MatMult.java
 *
 * A simple parallel matrix-matrix multiply example using jpvm.
 *
 * Adam J Ferrari
 * Tue Jun  4 10:31:02 EDT 1996
 *
 * Copyright (C) 1996  Adam J Ferrari
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
 * MA 02139, USA.
 */

import java.util.Date;

import jpvm.*;

class MatMult
{
	static final int		ParamTag			= 11;
	static final int		DoneTag				= 22;
	static final int		PipeTag				= 33;
	static final int		RollTag				= 44;
	static int				numTasks			= 0;
	static int				matDim				= 0;
	static int				taskMeshDim			= 0;
	static int				localPartitionDim	= 0;
	static int				localPartitionSize	= 0;
	static int				taskMeshRow			= 0;
	static int				taskMeshCol			= 0;
	static int				localTaskIndex		= 0;
	static boolean			debug				= false;
	static JPVMTaskId		myTaskId			= null;
	static JPVMTaskId		masterTaskId		= null;
	static JPVMEnvironment	jpvm				= null;
	static JPVMTaskId		tids[];
	static float			C[], A[], B[], tempA[];

	public static void main(String args[])
	{
		double mmstart = 0.0;
		double start = 0.0;
		double end = 0.0;
		boolean cmd = false;

		try
		{
			jpvm = new JPVMEnvironment();
			myTaskId = jpvm.pvm_mytid();
			masterTaskId = jpvm.pvm_parent();
			if (masterTaskId == JPVMEnvironment.PvmNoParent)
			{
				if (args.length != 2)
					usage();
				cmd = true;
				try
				{
					numTasks = Integer.parseInt(args[0]);
					matDim = Integer.parseInt(args[1]);
				}
				catch (NumberFormatException e)
				{
					usage();
				}
				System.out.println("" + matDim + "x" + matDim
						+ " matrix multiply, " + numTasks + " tasks");

				start = msecond();
				taskMeshDim = getTaskMeshDim(numTasks);

				tids = new JPVMTaskId[numTasks];
				if (numTasks > 1)
					jpvm.pvm_spawn("examples.MatMult", numTasks - 1, tids);
				tids[numTasks - 1] = tids[0];
				tids[0] = myTaskId;
				localTaskIndex = 0;

				/* Broadcast parameters to all tasks */
				mmstart = msecond();
				JPVMBuffer buf = new JPVMBuffer();
				buf.pack(taskMeshDim);
				buf.pack(numTasks);
				buf.pack(tids, numTasks, 1);
				buf.pack(matDim);
				jpvm.pvm_mcast(buf, tids, numTasks, ParamTag);
			}
			else
			{
				// Normal worker task - get parameters...
				JPVMMessage m = jpvm.pvm_recv(ParamTag);
				taskMeshDim = m.getBuffer().upkint();
				numTasks = m.getBuffer().upkint();
				tids = new JPVMTaskId[numTasks];
				m.getBuffer().unpack(tids, numTasks, 1);
				for (localTaskIndex = 0; localTaskIndex < numTasks; localTaskIndex++)
					if (myTaskId.equals(tids[localTaskIndex]))
						break;
				matDim = m.getBuffer().upkint();
			}

			/* Do the matrix multiplication */
			matmul();

			if (cmd)
			{
				end = msecond();
				System.out.println("Total time: " + (end - start) + " (mult: "
						+ (end - mmstart) + ")");
			}
			jpvm.pvm_exit();
		}
		catch (JPVMException jpe)
		{
			error("jpvm Exception - " + jpe.toString(), true);
		}
	}

	public static void matmul() throws JPVMException
	{
		int i, j, k;

		localPartitionDim = matDim / taskMeshDim;
		localPartitionSize = localPartitionDim * localPartitionDim;
		taskMeshRow = localTaskIndex / taskMeshDim;
		taskMeshCol = localTaskIndex % taskMeshDim;
		A = new float[localPartitionSize];
		B = new float[localPartitionSize];
		C = new float[localPartitionSize];
		tempA = new float[localPartitionSize];

		if (debug)
		{
			System.out.println("localTaskIndex\t= " + localTaskIndex);
			System.out.println("matDim\t= " + matDim);
			System.out.println("taskMeshDim\t= " + taskMeshDim);
			System.out.println("localPartitionDim\t= " + localPartitionDim);
			System.out.println("localPartitionSize\t= " + localPartitionSize);
			System.out.println("taskMeshRow\t= " + taskMeshRow);
			System.out.println("taskMeshCol\t= " + taskMeshCol);
		}

		for (i = 0; i < localPartitionSize; i++)
		{
			C[i] = (float) 0.0;
			A[i] = (float) (i + localTaskIndex * (taskMeshRow + 1));
			B[i] = (float) (i - localTaskIndex * (taskMeshRow + 1));
		}

		for (i = 0; i < taskMeshDim; i++)
		{
			Pipe(i);
			Multiply();
			if (i < (taskMeshDim - 1))
				Roll();
		}
	}

	public static void Pipe(int iter) throws JPVMException
	{
		int i;
		if (taskMeshCol == (taskMeshRow + iter) % taskMeshDim)
		{
			JPVMBuffer buf = new JPVMBuffer();
			buf.pack(A, localPartitionSize, 1);
			for (i = 0; i < taskMeshDim; i++)
				if (localTaskIndex != taskMeshRow * taskMeshDim + i)
				{
					jpvm.pvm_send(buf, tids[taskMeshRow * taskMeshDim + i],
							PipeTag);
				}
		}
		else
		{
			JPVMMessage m = jpvm.pvm_recv(PipeTag);
			m.getBuffer().unpack(tempA, localPartitionSize, 1);
		}
	}

	public static void Multiply() throws JPVMException
	{
		int i, j, k;
		float temp;
		for (i = 0; i < localPartitionDim; i++)
			for (j = 0; j < localPartitionDim; j++)
			{
				temp = 0;
				for (k = 0; k < localPartitionDim; k++)
					temp += A[i * localPartitionDim + k]
							* B[k * localPartitionDim + j];
				C[i * localPartitionDim + j] += temp;
			}
	}

	public static void Roll() throws JPVMException
	{
		int who = ((taskMeshRow != 0) ? (taskMeshRow - 1) * taskMeshDim
				+ taskMeshCol : (taskMeshDim - 1) * taskMeshDim + taskMeshCol);
		JPVMBuffer buf = new JPVMBuffer();
		buf.pack(B, localPartitionSize, 1);
		jpvm.pvm_send(buf, tids[who], RollTag);
		JPVMMessage m = jpvm.pvm_recv(RollTag);
		m.getBuffer().unpack(B, localPartitionSize, 1);
	}

	public static double msecond()
	{
		Date d = new Date();
		double msec = (double) d.getTime();
		return msec;
	}

	public static void error(String message, boolean die)
	{
		System.err.println("MatMult: " + message);
		if (die)
		{
			if (jpvm != null)
			{
				try
				{
					jpvm.pvm_exit();
				}
				catch (JPVMException jpe)
				{
				}
				System.exit(1);
			}
		}
	}

	public static void usage()
	{
		error("usage -  java MatMult <tasks> <mat dim>", true);
	}

	public static int getTaskMeshDim(int n)
	{
		int dim;
		for (dim = 1; dim <= n && n > 0; dim++)
		{
			if (dim * dim == n)
				return dim;
		}
		error("Number of tasks must be an even square", true);
		return -1;
	}
};
