/* JPVMDaemon.java
 *
 * The jpvm Daemon is a special jpvm task that runs on every 
 * host in a jpvm parallel virtual machine. The jpvm Daemon
 * is responsible for servicing requests to create new jpvm
 * tasks. Unlike standard PVM, the jpvm Daemon is never
 * responsible for routing message communication. 
 *
 * Adam J Ferrari
 * Sun 05-26-1996
 *
 * Copyright (C) 1996  Adam J Ferrari
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
 * MA 02139, USA.
 */

package jpvm;

import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;

public class JPVMDaemon
{
	private static JPVMEnvironment			jpvm			= null;
	private static JPVMTaskId				myTid			= null;

	private static JPVMTaskList				tasks			= null;
	private static JPVMTaskList				hosts			= null;
	private static JPVMSpawnWorkOrderList	spawnOrders		= null;
	private static int						maxCreateOrders	= 256;
	private static JPVMCreateWorkOrder		createOrders[];
	private static int						nextCreateOrder	= 0;

	private static boolean					log_on			= true;
	private static boolean					debug_on		= true;
	private static String					my_host_name	= null;

	private final static String SPACE = " ";

	public static void main(String args[])
	{
		try
		{
			int i;
			// Initialize data structures
			jpvm = new JPVMEnvironment(true);
			myTid = jpvm.pvm_mytid();
			tasks = new JPVMTaskList();
			hosts = new JPVMTaskList();
			spawnOrders = new JPVMSpawnWorkOrderList();
			createOrders = new JPVMCreateWorkOrder[maxCreateOrders];

			// Announce location
			log(jpvm.pvm_mytid().toString());
			my_host_name = jpvm.pvm_mytid().getHost();
			hosts.addTask(jpvm.pvm_mytid(), my_host_name);

			writeDaemonFile();

			// Main server loop...
			while (true)
			{
				JPVMMessage req = jpvm.pvm_recv();
				JPVMTaskId client = req.getSourceTid();
				int request = req.getMessageTag();
				String reqName;

				switch (request)
				{
					case (JPVMDaemonMessageTag.jpvmdPingRequest):
						reqName = "Ping";
						break;
					case (JPVMDaemonMessageTag.jpvmdRegisterTask):
						reqName = "RegisterTask";
						break;
					case (JPVMDaemonMessageTag.jpvmdRegisterChild):
						reqName = "RegisterChild";
						break;
					case JPVMDaemonMessageTag.jpvmdSpawnTask:
						reqName = "SpawnTask";
						break;
					case JPVMDaemonMessageTag.jpvmdCreateTask:
						reqName = "CreateTask";
						break;
					case JPVMDaemonMessageTag.jpvmdCreatedTask:
						reqName = "CreateTaskReturn";
						break;
					case JPVMDaemonMessageTag.jpvmdDeleteTask:
						reqName = "DeleteTask";
						break;
					case JPVMDaemonMessageTag.jpvmdTaskStatus:
						reqName = "TaskStatus";
						break;
					case JPVMDaemonMessageTag.jpvmdAddHost:
						reqName = "AddHost";
						break;
					case JPVMDaemonMessageTag.jpvmdAddHostBcast:
						reqName = "AddHostNotify";
						break;
					case JPVMDaemonMessageTag.jpvmdDeleteHost:
						reqName = "DeleteHost";
						break;
					case JPVMDaemonMessageTag.jpvmdDeleteHostBcast:
						reqName = "DeleteHostNotify";
						break;
					case JPVMDaemonMessageTag.jpvmdHostStatus:
						reqName = "HostStatus";
						break;
					case JPVMDaemonMessageTag.jpvmdHostHalt:
						reqName = "HostHalt";
						break;
					case JPVMDaemonMessageTag.jpvmdHalt:
						reqName = "Halt";
						break;
					default:
						reqName = "Unknown Request";
				}

				if (debug_on)
				{
					log("new request type=" + request + ",\"" + reqName
							+ "\", from " + client.toString());
				}

				switch (request)
				{
					case (JPVMDaemonMessageTag.jpvmdPingRequest):
						Ping(client, req.getBuffer());
						break;
					case (JPVMDaemonMessageTag.jpvmdRegisterTask):
						RegisterTask(client, req.getBuffer());
						break;
					case (JPVMDaemonMessageTag.jpvmdRegisterChild):
						RegisterChild(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdSpawnTask:
						SpawnTask(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdCreateTask:
						CreateTask(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdCreatedTask:
						CreatedTask(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdDeleteTask:
						DeleteTask(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdTaskStatus:
						TaskStatus(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdAddHost:
						AddHost(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdAddHostBcast:
						AddHostBcast(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdDeleteHost:
						DeleteHost(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdDeleteHostBcast:
						DeleteHostBcast(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdHostStatus:
						HostStatus(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdHostHalt:
						HostHalt(client, req.getBuffer());
						break;
					case JPVMDaemonMessageTag.jpvmdHalt:
						Halt(client, req.getBuffer());
						break;
					default:
						perror("unknown request type");
				}

			}
		}
		catch (JPVMException jpe)
		{
			JPVMDebug.note("JPVMDaemon, internal jpvm error.");
		}
	}
	
	private static String getBasePath()
	{
		try
		{
			URL url = JPVMDaemon.class.getResource("..");
		
			File file = new File(url.toURI());
			
			return(file.getAbsolutePath());
		}
		catch(Exception e)
		{
			e.printStackTrace();
			return new String();
		}
	}

	private static void daemonBcast(JPVMBuffer buf, int tag)
	{
		jpvmTaskListRecord tmp = hosts.firstIter();
		while (tmp != null)
		{
			try
			{
				jpvm.pvm_send(buf, tmp.tid, tag);
				tmp = hosts.nextIter();
			}
			catch (JPVMException jpe)
			{
				perror("problem sending to daemon " + tmp.tid.toString());
				hosts.deleteTask(tmp.tid);
			}
		}
	}

	private static void RegisterTask(JPVMTaskId client, JPVMBuffer req)
	{
		try
		{
			String name = req.upkstr();
			tasks.addTask(client, name);
		}
		catch (JPVMException jpe)
		{
			log("bad RegisterTask invocation");
		}
	}

	private static void RegisterChild(JPVMTaskId client, JPVMBuffer req)
	{
		int regNum = -1;

		// Child process reporting in. Notify the remote client that
		// requested this local task creation.

		try
		{
			regNum = req.upkint();
		}
		catch (JPVMException jpe)
		{
			perror("bad RegisterChild invocation");
			return;
		}

		if (regNum < 0 || regNum >= maxCreateOrders)
		{
			perror("RegisterChild, child registration number " + regNum
					+ "out of bounds");
			return;
		}

		JPVMCreateWorkOrder order = createOrders[regNum];
		if (order == null)
		{
			perror("RegisterChild, child registration number " + regNum
					+ "not expected");
			return;
		}
		if (!order.outstanding)
		{
			perror("RegisterChild, child registration number " + regNum
					+ "unexpected");
			return;
		}
		order.outstanding = false;

		// Return the blessed event to the requester
		try
		{
			JPVMBuffer buf = new JPVMBuffer();
			buf.pack(1);
			buf.pack(client);
			buf.pack(order.order);
			jpvm.pvm_send(buf, order.client,
					JPVMDaemonMessageTag.jpvmdCreatedTask);
		}
		catch (JPVMException jpe)
		{
			perror("RegisterChild, \"" + jpe + "\" sending to client "
					+ order.client.toString());
		}
	}

	private static void SpawnTask(JPVMTaskId client, JPVMBuffer req)
	{
		int num = 0;
		String name = null;
		JPVMBuffer buf = new JPVMBuffer();
		try
		{
			num = req.upkint();
			name = req.upkstr();
		}
		catch (JPVMException jpe)
		{
			perror("bad SpawnTask invocation");
			return;
		}
		if (num == 0)
		{
			buf.pack(num);
			try
			{
				jpvm.pvm_send(buf, client, JPVMDaemonMessageTag.jpvmdSpawnTask);
			}
			catch (JPVMException jpe)
			{
				perror("SpawnTask, problem sending " + "to client "
						+ client.toString());
			}
			return;
		}

		// Create a work order for the spawn
		JPVMSpawnWorkOrder order;
		order = spawnOrders.newOrder(num);
		order.client = client;

		// Create a request to create a task on a remote host
		JPVMBuffer creq = new JPVMBuffer();
		creq.pack(name); // Pack the class to create
		creq.pack(client); // Pack parent of the created tasks
		creq.pack(order.order);

		// Schedule on known hosts round robin style
		jpvmTaskListRecord target = null;
		for (int i = 0; i < num; i++)
		{
			if (target == null)
				target = hosts.firstIter();
			if (target == null)
			{
				perror("no hosts in SpawnTask invocation");
				return;
			}
			if (target.tid.equals(jpvm.pvm_mytid()))
			{
				creq.rewind();
				CreateTask(jpvm.pvm_mytid(), creq);
			}
			else
			{
				try
				{
					jpvm.pvm_send(creq, target.tid,
							JPVMDaemonMessageTag.jpvmdCreateTask);
				}
				catch (JPVMException jpe)
				{
					perror("SpawnTask, error scheduling " + "on host "
							+ target.tid.toString());
				}
			}
			target = hosts.nextIter();
		}
	}

	private static void CreatedTask(JPVMTaskId client, JPVMBuffer req)
	{
		int count = -1;
		JPVMTaskId child = null;
		int orderNum = -1;

		try
		{
			count = req.upkint();
			if (count == 1)
			{
				child = req.upktid();
				orderNum = req.upkint();
			}
		}
		catch (JPVMException jpe)
		{
			perror("CreatedTask, bad report from " + client.toString());
		}

		// Look up which spawn order this is in regards to
		JPVMSpawnWorkOrder order;
		order = spawnOrders.lookup(orderNum);

		if (order == null)
		{
			perror("CreatedTask, order number " + orderNum + " is not valid");
			return;
		}

		// Update the status of the order
		order.tids[order.numDone] = child;
		order.numDone++;

		if (order.numDone == order.num)
		{
			// The order is complete - return the good
			// news to the original client
			JPVMBuffer buf = new JPVMBuffer();
			buf.pack(order.numDone);
			buf.pack(order.tids, order.numDone, 1);
			try
			{
				jpvm.pvm_send(buf, order.client,
						JPVMDaemonMessageTag.jpvmdSpawnTask);
			}
			catch (JPVMException jpe)
			{
				perror("CreatedTask, \"" + jpe + "\" sending to client "
						+ order.client.toString());
			}

			// Throw away the order
			spawnOrders.doneOrder(order);
		}
	}

	private static void CreateTask(JPVMTaskId client, JPVMBuffer req)
	{
		String name = null;
		JPVMTaskId parent = null;
		int order;

		try
		{
			name = req.upkstr();
			parent = req.upktid();
			order = req.upkint();
		}
		catch (JPVMException jpe)
		{
			perror("bad CreateTask invocation");
			return;
		}

		if (createOrders[nextCreateOrder] == null)
			createOrders[nextCreateOrder] = new JPVMCreateWorkOrder();

		if (createOrders[nextCreateOrder].outstanding)
		{
			perror("too many outstanding task creation requests");
			return;
		}

		// Log the task creation request so when the task reports
		// in we'll know it was expected
		createOrders[nextCreateOrder].order = order;
		createOrders[nextCreateOrder].client = client;
		createOrders[nextCreateOrder].outstanding = true;
		
		// Create a thread to execute the new task
		
		StringBuffer buffer = new StringBuffer();
		buffer.append("java -cp ");
		buffer.append(getBasePath());
		buffer.append(" -Djpvm.daemon=");
		buffer.append(jpvm.pvm_mytid().getPort());
		buffer.append(" -Djpvm.parhost=");
		buffer.append(parent.getHost());
		buffer.append(" -Djpvm.parport=");
		buffer.append(parent.getPort());
		buffer.append(" -Djpvm.taskname=");
		buffer.append(name);
		buffer.append(" -Djpvm.regnum=");
		buffer.append(nextCreateOrder);
		buffer.append(SPACE);
		buffer.append(name);
		
		String argumentString = buffer.toString();
		
		if (debug_on)
			log("exec( " + argumentString + ")");

		JPVMExecTaskThread spawnThread;
		spawnThread = new JPVMExecTaskThread(jpvm, client, argumentString, createOrders[nextCreateOrder]);
		nextCreateOrder++;
		spawnThread.start();
	}

	private static void DeleteTask(JPVMTaskId client, JPVMBuffer req)
	{
		tasks.deleteTask(client);
	}

	private static void TaskStatus(JPVMTaskId client, JPVMBuffer req)
	{
		JPVMBuffer buf = new JPVMBuffer();
		int n = tasks.numTasks();
		JPVMTaskId tids[] = new JPVMTaskId[n];
		buf.pack(n);
		jpvmTaskListRecord tmp = tasks.firstIter();
		int i = 0;
		while (tmp != null)
		{
			tids[i] = tmp.tid;
			buf.pack(tmp.name);
			tmp = tasks.nextIter();
			i++;
		}
		buf.pack(tids, n, 1);
		try
		{
			jpvm.pvm_send(buf, client, JPVMDaemonMessageTag.jpvmdTaskStatus);
		}
		catch (JPVMException jpe)
		{
			perror("TaskStatus, \"" + jpe + "\" sending to client "
					+ client.toString());
		}
	}

	private static void AddHost(JPVMTaskId client, JPVMBuffer req)
	{
		JPVMBuffer buf = internalAddHosts(req);
		if (buf == null)
		{
			perror("AddHost, problem adding hosts");
			return;
		}
		daemonBcast(buf, JPVMDaemonMessageTag.jpvmdAddHostBcast);
	}

	private static void Ping(JPVMTaskId client, JPVMBuffer req)
	{
		try
		{
			jpvm.pvm_send(req, client, JPVMDaemonMessageTag.jpvmdPingReply);
		}
		catch (JPVMException jpe)
		{
			perror("ping, \"" + jpe + "\" sending to client "
					+ client.toString());
		}
	}

	private static void AddHostBcast(JPVMTaskId client, JPVMBuffer req)
	{
		if (client.equals(myTid))
			return;
		try
		{
			int i;
			int num = req.upkint();
			String names[] = new String[num];
			JPVMTaskId daemonTids[] = new JPVMTaskId[num];
			for (i = 0; i < num; i++)
				names[i] = req.upkstr();
			req.unpack(daemonTids, num, 1);
			for (i = 0; i < num; i++)
				hosts.addTask(daemonTids[i], names[i]);
		}
		catch (JPVMException jpe)
		{
			log("bad AddHost invocation");
		}
	}

	private static JPVMBuffer internalAddHosts(JPVMBuffer req)
	{
		int i, j;
		JPVMBuffer ret = new JPVMBuffer();

		int newNum = 0;
		String newNames[] = null;
		JPVMTaskId newTids[] = null;
		boolean newValid[] = null;
		int newValidNum = 0;
		try
		{
			// First, get the addresses of all new daemons
			newNum = req.upkint();
			newNames = new String[newNum];
			newTids = new JPVMTaskId[newNum];
			newValid = new boolean[newNum];
			for (i = 0; i < newNum; i++)
				newNames[i] = req.upkstr();
			req.unpack(newTids, newNum, 1);
		}
		catch (JPVMException jpe)
		{
			log("bad AddHost call");
			perror("internalAddHost, " + jpe);
			return null;
		}

		// Check the validity of all new names
		JPVMBuffer pingBuf = new JPVMBuffer();
		newValidNum = newNum;
		for (i = 0; i < newNum; i++)
		{
			boolean valid = true;
			try
			{
				jpvm.pvm_send(pingBuf, newTids[i],
						JPVMDaemonMessageTag.jpvmdPingRequest);
				JPVMMessage pingMess = jpvm.pvm_recv(newTids[i],
						JPVMDaemonMessageTag.jpvmdPingReply);
			}
			catch (JPVMException jpe)
			{
				valid = false;
				newValidNum--;
				perror("internalAddHost, ping, " + jpe);
			}
			newValid[i] = valid;
			if (valid)
				hosts.addTask(newTids[i], newNames[i]);
		}

		if (newValidNum < 1)
		{
			// no hosts to add!
			ret = null;
			perror("internalAddHost, no hosts added");
			return ret;
		}

		// Create the message to add all daemons, new and old
		int oldNum = hosts.numTasks();
		int totalNum = oldNum + newValidNum;
		JPVMTaskId allTids[] = new JPVMTaskId[totalNum];
		jpvmTaskListRecord tmp = hosts.firstIter();

		// Pack in the old names...
		ret.pack(totalNum);
		i = 0;
		while (tmp != null)
		{
			allTids[i] = tmp.tid;
			ret.pack(tmp.name);
			tmp = hosts.nextIter();
			i++;
		}
		// Pack in the old names...
		for (j = 0; j < newNum; j++)
			if (newValid[j])
			{
				allTids[i] = newTids[j];
				ret.pack(newNames[j]);
				i++;
			}

		// Pack in all of the tids...
		ret.pack(allTids, totalNum, 1);
		return ret;
	}

	private static void DeleteHost(JPVMTaskId client, JPVMBuffer req)
	{
	}

	private static void DeleteHostBcast(JPVMTaskId client, JPVMBuffer req)
	{
	}

	private static void HostStatus(JPVMTaskId client, JPVMBuffer req)
	{
		JPVMBuffer buf = new JPVMBuffer();
		int nhosts = hosts.numTasks();
		buf.pack(nhosts);
		JPVMTaskId dTids[] = new JPVMTaskId[nhosts];
		jpvmTaskListRecord tmp = hosts.firstIter();
		int i = 0;
		while (tmp != null)
		{
			dTids[i] = tmp.tid;
			buf.pack(tmp.name);
			tmp = hosts.nextIter();
			i++;
		}
		buf.pack(dTids, nhosts, 1);
		try
		{
			jpvm.pvm_send(buf, client, JPVMDaemonMessageTag.jpvmdHostStatus);
		}
		catch (JPVMException jpe)
		{
			perror("HostStatus, \"" + jpe + "\" sending to client "
					+ client.toString());
		}
	}

	private static void HostHalt(JPVMTaskId client, JPVMBuffer req)
	{
		log("shutting down");
		System.exit(0);
	}

	private static void Halt(JPVMTaskId client, JPVMBuffer req)
	{
		JPVMBuffer buf = new JPVMBuffer();
		daemonBcast(buf, JPVMDaemonMessageTag.jpvmdHostHalt);
	}

	private static void log(String message)
	{
		if (log_on)
		{
			System.out.println("jpvm daemon: " + message);
			System.out.flush();
		}
	}

	private static void perror(String message)
	{
		System.err.println("jpvm daemon: " + message);
		System.err.flush();
	}

	private static void writeDaemonFile()
	{
		String fileName = JPVMEnvironment.pvm_daemon_file_name();
		try
		{
			File f = new File(fileName);
			FileOutputStream fout = new FileOutputStream(f);
			DataOutputStream dout = new DataOutputStream(fout);
			int port = myTid.getPort();
			dout.writeInt(port);
			dout.flush();
		}
		catch (IOException ioe)
		{
			perror("error writing \"" + fileName + "\"");
		}
	}
};

class jpvmTaskListRecord
{
	public JPVMTaskId			tid;
	public String				name;
	public jpvmTaskListRecord	next;

	public jpvmTaskListRecord(JPVMTaskId t, String n)
	{
		tid = t;
		name = n;
		next = null;
	}
};

class JPVMTaskList
{
	jpvmTaskListRecord	tasks;
	int					num_tasks;
	jpvmTaskListRecord	iter;

	public JPVMTaskList()
	{
		tasks = null;
		num_tasks = 0;
		iter = null;
	}

	public int numTasks()
	{
		return num_tasks;
	}

	public void addTask(JPVMTaskId tid, String name)
	{
		if (find(tid) != null)
		{
			// Already know about this task...
			return;
		}
		jpvmTaskListRecord nw = new jpvmTaskListRecord(tid, name);
		nw.next = tasks;
		tasks = nw;
		num_tasks++;
	}

	public void deleteTask(JPVMTaskId tid)
	{
		if (tasks == null)
			return;
		jpvmTaskListRecord tmp = tasks;

		// Check head
		if (tmp.tid.equals(tid))
		{
			if (iter == tmp)
				iter = tmp.next;
			tasks = tasks.next;
			num_tasks--;
			return;
		}
		// Check body
		while (tmp.next != null)
		{
			if (tmp.next.tid.equals(tid))
			{
				if (iter == tmp.next)
					iter = tmp.next.next;
				tmp.next = tmp.next.next;
				num_tasks--;
				return;
			}
			tmp = tmp.next;
		}
	}

	public jpvmTaskListRecord find(JPVMTaskId tid)
	{
		jpvmTaskListRecord tmp = tasks;
		while (tmp != null)
		{
			if (tmp.tid.equals(tid))
				return tmp;
			tmp = tmp.next;
		}
		return tmp;
	}

	public jpvmTaskListRecord firstIter()
	{
		if (tasks == null)
			return null;
		jpvmTaskListRecord ret = tasks;
		iter = tasks.next;
		return ret;
	}

	public jpvmTaskListRecord nextIter()
	{
		if (iter == null)
			return null;
		jpvmTaskListRecord ret = iter;
		iter = iter.next;
		return ret;
	}
};

class JPVMCreateWorkOrder
{
	public int			order;
	public JPVMTaskId	client;
	public boolean		outstanding;

	JPVMCreateWorkOrder()
	{
		order = 0;
		client = null;
		outstanding = false;
	}
};

class JPVMSpawnWorkOrder
{
	public int			order;		// Which partially completed spawn
	public JPVMTaskId	tids[];	// Tids spawned
	public int			num;		// Number to spawn
	public int			numDone;	// Number actually done
	public JPVMTaskId	client;	// Who placed the order?
	JPVMSpawnWorkOrder	next;		// Linked list

	JPVMSpawnWorkOrder(int o, int n)
	{
		order = 0;
		num = n;
		tids = new JPVMTaskId[n];
		numDone = 0;
		client = null;
		next = null;
	}
};

class JPVMSpawnWorkOrderList
{
	private JPVMSpawnWorkOrder	list		= null;
	private int					nextOrder	= 1;

	JPVMSpawnWorkOrderList()
	{
		list = null;
		nextOrder = 1;
	}

	public JPVMSpawnWorkOrder newOrder(int num)
	{
		JPVMSpawnWorkOrder ret;

		ret = new JPVMSpawnWorkOrder(nextOrder, num);
		nextOrder++;
		ret.next = list;
		list = ret;
		return ret;
	}

	public JPVMSpawnWorkOrder lookup(int order)
	{
		JPVMSpawnWorkOrder ret;

		ret = list;

		while (ret != null)
		{
			if (ret.order == order)
				return ret;
			ret = ret.next;
		}
		return null;
	}

	public void doneOrder(JPVMSpawnWorkOrder order)
	{
		JPVMSpawnWorkOrder tmp;
		if (list == null || order == null)
			return;
		if (order == list)
		{
			list = list.next;
			return;
		}
		tmp = list;
		while (tmp.next != null)
		{
			if (tmp.next == order)
			{
				tmp.next = order.next;
				return;
			}
			tmp = tmp.next;
		}
	}

	public void doneOrder(int order)
	{
		JPVMSpawnWorkOrder tmp;
		if (list == null)
			return;
		if (order == list.order)
		{
			list = list.next;
			return;
		}
		tmp = list;
		while (tmp.next != null)
		{
			if (tmp.next.order == order)
			{
				tmp.next = tmp.next.next;
				return;
			}
			tmp = tmp.next;
		}
	}
};

class JPVMDaemonWaiter extends Thread
{
	private Process	process;

	JPVMDaemonWaiter(Process p)
	{
		process = p;
	}

	public void run()
	{
		boolean wait = true;
		while (wait)
		{
			try
			{
				process.waitFor();
				wait = false;
			}
			catch (InterruptedException ie)
			{
			}
		}
	}
};

class JPVMExecTaskThread extends Thread
{
	private JPVMEnvironment		jpvm		= null;
	private Process				process;
	private JPVMTaskId			client;
	private String				argumentString;
	private boolean				log_on		= true;
	private boolean				debug_on	= false;
	private JPVMCreateWorkOrder	order;

	JPVMExecTaskThread(JPVMEnvironment j, JPVMTaskId c, String argumentString,	JPVMCreateWorkOrder o)
	{
		jpvm = j;
		client = c;
		this.argumentString = argumentString;
		order = o;
	}

	private boolean doExec()
	{
		try
		{
			process = Runtime.getRuntime().exec(argumentString);//java -cp D:/Workspace/SysSoft_bungen/src -Djpvm.daemon=2022 -Djpvm.parhost=glenmorangie -Djpvm.parport=2023 -Djpvm.taskname=examples.HelloOther -Djpvm.regnum=0 examples.HelloOther");
		}
		catch (IOException ioe)
		{
			perror("i/o exception on exec");
			order.outstanding = false;
			JPVMBuffer buf = new JPVMBuffer();
			buf.pack(order.order);
			buf.pack(-1);
			try
			{
				jpvm.pvm_send(buf, client,
						JPVMDaemonMessageTag.jpvmdCreatedTask);
			}
			catch (JPVMException jpe)
			{
				perror("CreateTask, \"" + jpe + " sending " + "to client "
						+ client.toString());
			}
			return false;
		}
		return true;
	}

	public void run()
	{
		boolean wait = doExec();
		while (wait)
		{
			try
			{
				process.waitFor();
				wait = false;
			}
			catch (InterruptedException ie)
			{
			}
		}
	}

	private void log(String message)
	{
		if (log_on)
		{
			System.out.println("jpvm daemon: " + message);
			System.out.flush();
		}
	}

	private void perror(String message)
	{
		System.err.println("jpvm daemon: " + message);
		System.err.flush();
	}
};
