/* JPVMEnvironment.java
 *
 * The JPVMEnvironment class implements the jpvm run time environment
 * of a task enrolled in the jpvm parallel virtual machine.
 *
 * Adam J Ferrari
 * Sat 05-25-1996
 *
 * Copyright (C) 1996  Adam J Ferrari
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
 * MA 02139, USA.
 */

package jpvm;

import jpvm.JPVMTaskId;
import jpvm.JPVMMessage;
import jpvm.JPVMMessageQueue;
import jpvm.JPVMConnectionSet;
import jpvm.JPVMConnectionServer;
import jpvm.JPVMConfiguration;
import jpvm.JPVMTaskStatus;
import java.io.*;

public class JPVMEnvironment
{
	public static final JPVMTaskId		PvmNoParent			= null;

	private static JPVMTaskId			myTid;
	private static JPVMTaskId			parentTid;
	private static JPVMTaskId			daemonTid;
	private static JPVMMessageQueue		myMessageQueue;
	private static JPVMConnectionSet	myConnectionSet;
	private static JPVMConnectionServer	connectionServer;
	private static String				myName;
	private static int					registrationNumber	= -1;

	public JPVMEnvironment() throws JPVMException
	{
		init(false, null);
	}

	public JPVMEnvironment(String taskName) throws JPVMException
	{
		init(false, taskName);
	}

	public JPVMEnvironment(boolean isDaemon) throws JPVMException
	{
		init(isDaemon, null);
	}

	public JPVMEnvironment(boolean isDaemon, String taskName)
			throws JPVMException
	{
		init(isDaemon, taskName);
	}

	public JPVMTaskId pvm_mytid()
	{
		return myTid;
	}

	public JPVMTaskId pvm_parent()
	{
		return parentTid;
	}

	public int pvm_spawn(String task_name, int num, JPVMTaskId tids[]) throws JPVMException
	{
		int ret = 0;
		JPVMBuffer buf = new JPVMBuffer();
		buf.pack(num);
		buf.pack(task_name);
		pvm_send(buf, daemonTid, JPVMDaemonMessageTag.jpvmdSpawnTask);
		try
		{
			JPVMMessage m = pvm_recv(JPVMDaemonMessageTag.jpvmdSpawnTask);
			ret = m.getBuffer().upkint();
			m.getBuffer().unpack(tids, ret, 1);
		}
		catch (JPVMException jpe)
		{
			JPVMDebug.error("pvm_spawn, internal error");
		}
		return ret;
	}

	public synchronized void pvm_send(JPVMBuffer buf, JPVMTaskId tid, int messageTag)
			throws JPVMException
	{
		JPVMDebug.note("pvm_send, sending message to " + tid.toString());
		JPVMMessage message = new JPVMMessage(buf, tid, myTid, messageTag);
		if (myTid.equals(tid))
		{
			// Just enqueue the message
			myMessageQueue.enqueue(message);
		}
		else
		{
			JPVMSendConnection conn = getConnection(tid);
			message.send(conn);
		}
	}

	public synchronized void pvm_mcast(JPVMBuffer buf, JPVMTaskId tids[],
			int ntids, int tag) throws JPVMException
	{
		int exceptions = 0;
		JPVMMessage message = new JPVMMessage(buf, null, myTid, tag);
		for (int i = 0; i < ntids; i++)
		{
			JPVMTaskId tid = tids[i];
			message.setDestTid(tid);
			JPVMDebug.note("pvm_mcast, sending message to " + tid.toString());
			try
			{
				JPVMSendConnection conn = getConnection(tid);
				message.send(conn);
			}
			catch (JPVMException jpe)
			{
				exceptions++;
			}
		}
		if (exceptions > 0)
			throw new JPVMException("pvm_mcast, some messages " + "failed");
	}

	public JPVMMessage pvm_recv(JPVMTaskId tid, int tag) throws JPVMException
	{
		//
		// Thanks to Professor Thomas R. James of the Dept. of
		// Mathematical Sciences at Otterbein College for finding
		// and fixing race condition in the previous implementation
		// of pvm_recv.
		//
		// Adam Ferrari - Mon Feb 1 13:11:12 EST 1999
		//
		return myMessageQueue.dequeue(tid, tag);
	}

	public JPVMMessage pvm_recv(JPVMTaskId tid) throws JPVMException
	{
		return myMessageQueue.dequeue(tid);
	}

	public JPVMMessage pvm_recv(int tag) throws JPVMException
	{
		return myMessageQueue.dequeue(tag);
	}

	public JPVMMessage pvm_recv() throws JPVMException
	{
		return myMessageQueue.dequeue();
	}

	public JPVMMessage pvm_nrecv(JPVMTaskId tid, int tag) throws JPVMException
	{
		return myMessageQueue.dequeueNonBlock(tid, tag);
	}

	public JPVMMessage pvm_nrecv(JPVMTaskId tid) throws JPVMException
	{
		return myMessageQueue.dequeueNonBlock(tid);
	}

	public JPVMMessage pvm_nrecv(int tag) throws JPVMException
	{
		return myMessageQueue.dequeueNonBlock(tag);
	}

	public JPVMMessage pvm_nrecv() throws JPVMException
	{
		return myMessageQueue.dequeueNonBlock();
	}

	public boolean pvm_probe(JPVMTaskId tid, int tag) throws JPVMException
	{
		return myMessageQueue.probe(tid, tag);
	}

	public boolean pvm_probe(JPVMTaskId tid) throws JPVMException
	{
		return myMessageQueue.probe(tid);
	}

	public boolean pvm_probe(int tag) throws JPVMException
	{
		return myMessageQueue.probe(tag);
	}

	public boolean pvm_probe() throws JPVMException
	{
		return myMessageQueue.probe();
	}

	public void pvm_exit() throws JPVMException
	{
		JPVMBuffer buf = new JPVMBuffer();
		try
		{
			Thread.sleep(1000);
		}
		catch (InterruptedException ie)
		{
		}
		pvm_send(buf, daemonTid, JPVMDaemonMessageTag.jpvmdDeleteTask);
	}

	public JPVMConfiguration pvm_config()
	{
		JPVMBuffer buf = new JPVMBuffer();
		JPVMConfiguration ret = null;
		try
		{
			pvm_send(buf, daemonTid, JPVMDaemonMessageTag.jpvmdHostStatus);
			JPVMMessage m = pvm_recv(JPVMDaemonMessageTag.jpvmdHostStatus);
			int n = m.getBuffer().upkint();
			ret = new JPVMConfiguration(n);
			for (int i = 0; i < n; i++)
				ret.hostNames[i] = m.getBuffer().upkstr();
			m.getBuffer().unpack(ret.hostDaemonTids, n, 1);
		}
		catch (JPVMException jpe)
		{
			JPVMDebug.error("pvm_config, internal error");
		}
		return ret;
	}

	public JPVMTaskStatus pvm_tasks(JPVMConfiguration conf, int which)
	{
		JPVMTaskStatus ret = null;
		if (conf == null || which < 0 || which >= conf.numHosts)
			return null;

		try
		{
			JPVMBuffer buf = new JPVMBuffer();
			pvm_send(buf, conf.hostDaemonTids[which],
					JPVMDaemonMessageTag.jpvmdTaskStatus);
			JPVMMessage m = pvm_recv(JPVMDaemonMessageTag.jpvmdTaskStatus);
			ret = new JPVMTaskStatus();
			ret.hostName = conf.hostNames[which];
			ret.numTasks = m.getBuffer().upkint();
			if (ret.numTasks == 0)
			{
				ret.taskNames = null;
				ret.taskTids = null;
				return ret;
			}
			ret.taskNames = new String[ret.numTasks];
			ret.taskTids = new JPVMTaskId[ret.numTasks];
			for (int i = 0; i < ret.numTasks; i++)
				ret.taskNames[i] = m.getBuffer().upkstr();
			m.getBuffer().unpack(ret.taskTids, ret.numTasks, 1);
		}
		catch (JPVMException jpe)
		{
			JPVMDebug.error("pvm_tasks, internal error");
		}
		return ret;
	}

	public void pvm_halt() throws JPVMException
	{
		JPVMBuffer buf = new JPVMBuffer();
		pvm_send(buf, daemonTid, JPVMDaemonMessageTag.jpvmdHalt);
	}

	public void pvm_addhosts(int nhosts, String hostnames[],
			JPVMTaskId daemonTids[]) throws JPVMException
	{
		JPVMBuffer buf = new JPVMBuffer();
		buf.pack(nhosts);
		for (int i = 0; i < nhosts; i++)
			buf.pack(hostnames[i]);
		buf.pack(daemonTids, nhosts, 1);
		pvm_send(buf, daemonTid, JPVMDaemonMessageTag.jpvmdAddHost);
	}

	// Internal methods:
	private void init(boolean isDaemon, String taskName) throws JPVMException
	{
		myMessageQueue = new JPVMMessageQueue();
		myConnectionSet = new JPVMConnectionSet();
		connectionServer = new JPVMConnectionServer(myConnectionSet, myMessageQueue);
		myTid = new JPVMTaskId(connectionServer.getConnectionPort());
		connectionServer.setDaemon(true);
		connectionServer.start();
		if (!isDaemon)
		{
			findDaemon();
			findParent();
			registerDaemon(taskName);
		}
	}

	private JPVMTaskId initTaskId()
	{
		JPVMTaskId ret = null;
		return ret;
	}

	private JPVMSendConnection getConnection(JPVMTaskId tid)
			throws JPVMException
	{
		JPVMSendConnection ret = null;
		ret = myConnectionSet.lookupSendConnection(tid);
		if (ret != null)
		{
			// Had a cached connection...
			return ret;
		}
		// Must establish a new connection.
		ret = JPVMSendConnection.connect(tid, myTid);
		if (ret != null)
		{
			myConnectionSet.insertSendConnection(ret);
			return ret;
		}
		throw new JPVMException("getConnection, connect failed");
	}

	private void findDaemon()
	{
		int daemonPort = -1;
		String daemonPortStr = System.getProperty("jpvm.daemon");
		if (daemonPortStr != null)
		{
			try
			{
				daemonPort = Integer.valueOf(daemonPortStr).intValue();
			}
			catch (NumberFormatException nfe)
			{
				JPVMDebug.error("couldn't bind to daemon, "
						+ "jpvm.daemon not an integer");
				daemonPort = -1;
			}
		}
		else
		{
			daemonPort = readDaemonFile();
		}
		if (daemonPort == -1)
		{
			JPVMDebug.error("couldn't bind to daemon, "
					+ "jpvm.daemon not defined");
		}
		daemonTid = new JPVMTaskId(daemonPort);
	}

	private void findParent() throws JPVMException
	{
		String parentHost = System.getProperty("jpvm.parhost");
		int parentPort = 0;
		if (parentHost == null)
		{
			parentTid = null;
			return;
		}
		String parentPortStr = System.getProperty("jpvm.parport");
		try
		{
			parentPort = Integer.valueOf(parentPortStr).intValue();
		}
		catch (NumberFormatException nfe)
		{
			JPVMDebug.error("couldn't bind to parent, "
					+ "jpvm.parport not an integer");
		}
		parentTid = new JPVMTaskId(parentHost, parentPort);

		// Since we have a parent, register with the daemon
		String regStr = System.getProperty("jpvm.regnum");
		if (regStr == null)
		{
			JPVMDebug.error("no task registration number");
		}
		else
		{
			try
			{
				registrationNumber = Integer.valueOf(regStr).intValue();
			}
			catch (NumberFormatException nfe)
			{
				JPVMDebug.error("invalid task registration number");
			}
		}
		JPVMBuffer buf = new JPVMBuffer();
		buf.pack(registrationNumber);
		pvm_send(buf, daemonTid, JPVMDaemonMessageTag.jpvmdRegisterChild);
	}

	private void registerDaemon(String taskName) throws JPVMException
	{
		// Find out the name of this task
		if (taskName == null)
		{
			myName = System.getProperty("jpvm.taskname");
			if (myName == null)
				myName = "(command line jpvm task)";
		}
		else
		{
			myName = new String(taskName);
		}

		// Register this task with the daemon
		JPVMBuffer buf = new JPVMBuffer();
		buf.pack(myName);
		pvm_send(buf, daemonTid, JPVMDaemonMessageTag.jpvmdRegisterTask);
	}

	private int readDaemonFile()
	{
		int port = -1;
		String fileName = pvm_daemon_file_name();
		try
		{
			File f = new File(fileName);
			FileInputStream fin = new FileInputStream(f);
			DataInputStream din = new DataInputStream(fin);
			port = din.readInt();
		}
		catch (IOException ioe)
		{
			JPVMDebug.error("error reading \"" + fileName + "\"");
			port = -1;
		}
		return port;
	}

	public static String pvm_daemon_file_name()
	{
		String osName = System.getProperty("os.name");
		String userName = System.getProperty("user.name");
		String fileName = null;
		
		if (osName.equals("Windows 2000") || osName.equals("Windows XP"))
		{
			fileName = "c:\\temp\\jpvmd-" + userName + ".txt";
		}
		else
		{
			fileName = "/tmp/jpvmd." + userName;
		}
		return fileName;
	}
};
