/* JPVMMessageQueue.java
 *
 * The JPVMMessageQueue class implements a searchable database
 * of jpvm messages. The database can be searched by message tag,
 * source process, both, or neither (i.e. wildcard search).
 *
 * Adam J Ferrari
 * Sat 05-25-1996
 *
 * Copyright (C) 1996  Adam J Ferrari
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
 * MA 02139, USA.
 */

package jpvm;

class jpvmMessageQueueElement
{
	public JPVMMessage				message;
	public jpvmMessageQueueElement	next;

	public jpvmMessageQueueElement()
	{
		message = null;
		next = null;
	}

	public jpvmMessageQueueElement(JPVMMessage m)
	{
		message = m;
		next = null;
	}
};

public class JPVMMessageQueue
{
	private jpvmMessageQueueElement	list_head;
	private jpvmMessageQueueElement	list_tail;

	private synchronized void addElement(jpvmMessageQueueElement nw)
	{
		if (list_head == null)
		{
			list_head = list_tail = nw;
			return;
		}
		list_tail.next = nw;
		list_tail = nw;
	}

	private synchronized void deleteElement(jpvmMessageQueueElement d)
	{
		if (list_head == null)
			return;
		if (list_head == d)
		{
			// Deleting head element.
			list_head = d.next;
			if (list_tail == d)
				list_tail = null;
			return;
		}
		jpvmMessageQueueElement tmp = list_head;
		while (tmp.next != d)
		{
			tmp = tmp.next;
			if (tmp == null)
			{
				// Element wasn't in the list
				return;
			}
		}
		tmp.next = d.next;
		if (list_tail == d)
			list_tail = tmp;
	}

	private synchronized jpvmMessageQueueElement find()
	{
		return list_head;
	}

	private synchronized jpvmMessageQueueElement find(int tag)
	{
		jpvmMessageQueueElement tmp = list_head;
		while (tmp != null)
		{
			if (tmp.message.getMessageTag() == tag)
				return tmp;
			tmp = tmp.next;
		}
		return null;
	}

	private synchronized jpvmMessageQueueElement find(JPVMTaskId tid)
	{
		jpvmMessageQueueElement tmp = list_head;
		while (tmp != null)
		{
			if (tmp.message.getSourceTid().equals(tid))
				return tmp;
			tmp = tmp.next;
		}
		return null;
	}

	private synchronized jpvmMessageQueueElement find(JPVMTaskId tid, int tag)
	{
		jpvmMessageQueueElement tmp = list_head;
		while (tmp != null)
		{
			if ((tmp.message.getSourceTid().equals(tid))
					&& (tmp.message.getMessageTag() == tag))
				return tmp;
			tmp = tmp.next;
		}
		return null;
	}

	public JPVMMessageQueue()
	{
		list_head = list_tail = null;
	}

	public synchronized void enqueue(JPVMMessage m)
	{
		jpvmMessageQueueElement nw = new jpvmMessageQueueElement(m);
		addElement(nw);
		notifyAll();
	}

	public synchronized boolean probe()
	{
		jpvmMessageQueueElement tmp = find();
		return (tmp != null);
	}

	public synchronized boolean probe(int tag)
	{
		jpvmMessageQueueElement tmp = find(tag);
		return (tmp != null);
	}

	public synchronized boolean probe(JPVMTaskId tid)
	{
		jpvmMessageQueueElement tmp = find(tid);
		return (tmp != null);
	}

	public synchronized boolean probe(JPVMTaskId tid, int tag)
	{
		jpvmMessageQueueElement tmp = find(tid, tag);
		return (tmp != null);
	}

	public synchronized JPVMMessage dequeue()
	{
		jpvmMessageQueueElement tmp = null;
		while (true)
		{
			if ((tmp = find()) != null)
			{
				deleteElement(tmp);
				return tmp.message;
			}
			try
			{
				wait();
			}
			catch (InterruptedException ie)
			{
			}
		}
	}

	public synchronized JPVMMessage dequeue(int tag)
	{
		jpvmMessageQueueElement tmp = null;
		while (true)
		{
			if ((tmp = find(tag)) != null)
			{
				deleteElement(tmp);
				return tmp.message;
			}
			try
			{
				wait();
			}
			catch (InterruptedException ie)
			{
			}
		}
	}

	public synchronized JPVMMessage dequeue(JPVMTaskId tid)
	{
		jpvmMessageQueueElement tmp = null;
		while (true)
		{
			if ((tmp = find(tid)) != null)
			{
				deleteElement(tmp);
				return tmp.message;
			}
			try
			{
				wait();
			}
			catch (InterruptedException ie)
			{
			}
		}
	}

	public synchronized JPVMMessage dequeue(JPVMTaskId tid, int tag)
	{
		jpvmMessageQueueElement tmp = null;
		while (true)
		{
			if ((tmp = find(tid, tag)) != null)
			{
				deleteElement(tmp);
				return tmp.message;
			}
			try
			{
				wait();
			}
			catch (InterruptedException ie)
			{
			}
		}
	}

	public synchronized JPVMMessage dequeueNonBlock()
	{
		jpvmMessageQueueElement tmp = find();
		if (tmp != null)
		{
			deleteElement(tmp);
			return tmp.message;
		}
		return null;
	}

	public synchronized JPVMMessage dequeueNonBlock(int tag)
	{
		jpvmMessageQueueElement tmp = find(tag);
		if (tmp != null)
		{
			deleteElement(tmp);
			return tmp.message;
		}
		return null;
	}

	public synchronized JPVMMessage dequeueNonBlock(JPVMTaskId tid)
	{
		jpvmMessageQueueElement tmp = find(tid);
		if (tmp != null)
		{
			deleteElement(tmp);
			return tmp.message;
		}
		return null;
	}

	public synchronized JPVMMessage dequeueNonBlock(JPVMTaskId tid, int tag)
	{
		jpvmMessageQueueElement tmp = find(tid, tag);
		if (tmp != null)
		{
			deleteElement(tmp);
			return tmp.message;
		}
		return null;
	}
};
